/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <oskit/x86/pc/reset.h>
#include <oskit/x86/i16.h>
#include <oskit/x86/base_trap.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/pc/base_i16.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/error.h>
#include <oskit/exec/exec.h>
#include <oskit/lmm.h>
#include <oskit/machine/phys_lmm.h>
#include <oskit/diskpart/pcbios.h>
#include <oskit/boot/bootwhat.h>

#include "boot.h"
#include "pxe.h"
#include "decls.h"
#include "udp.h"

/*
 * For now, hardwired.
 */
#define NETBOOT		"/tftpboot/netboot"

/*
 * Defined in crt0.S
 */
extern unsigned int	pxenv_segment;		/* Segment of PXENV+ */
extern unsigned int	pxenv_offset;		/* Offset of PXENV+ */
extern SEGOFF16_t	pxe_entrypoint;		/* Entrypoint (16bit) of PXE */

/*
 * PXENV+ and !PXE pointers.
 */
static pxe_t		*pxep;
static pxenv_t		*pxenvp;

/*
 * Our DHCP information.
 */
struct in_addr		myip, serverip, gatewayip; 

/* Forward decls */
static int		boot_mbr_default(void);
static int		boot_mbr_sysid(int systemid);
static int		boot_mbr_partition(int partition);
static void		boot_system(void);
void			delay(int);
int			biosread(int, int, int, int, int, void *);

int
main(int argc, char **argv)
{
	BOOTPLAYER		bootplayer;
	t_PXENV_GET_CACHED_INFO cacheinfo;
		
	printf("\nPXE boot program (compiled %s)\n", __DATE__);

	/*
	 * Construct 32 bit addresses for these two.
	 */
	pxenvp = (pxenv_t *) ((pxenv_segment << 4) + pxenv_offset);
	pxep   = (pxe_t *) ((pxenvp->PXEPtr.segment << 4) +
			    pxenvp->PXEPtr.offset);	

	DPRINTF("PXENV+ at %p, !PXE at %p\n", pxenvp, pxep);

	/*
	 * Check that what we think is !PXE, really is.
	 */
	if (strncmp("!PXE", pxep->Signature, sizeof(pxep->Signature)))
		PANIC("Failed the \"!PXE\" test!");

	/*
	 * Suck out the 16bit PXE entrypoint and stash for the trampoline
	 * code to find for the intersegment call to the PXE.
	 */
	pxe_entrypoint = pxep->EntryPointSP;
	DPRINTF("PXE Entrypoint is 0x%x (offset:segment)\n",
		*(int *) &pxe_entrypoint);

	/*
	 * Get the information we need from the PXE to find the boot
	 * server. This is the response to the last DHCP request, which
	 * provided the name of this program.
	 */
	cacheinfo.Status         = 0;
	cacheinfo.PacketType     = PXENV_PACKET_TYPE_BINL_REPLY;
	cacheinfo.BufferSize     = sizeof(bootplayer);
	cacheinfo.Buffer.segment = (oskit_u16_t) 0;
	cacheinfo.Buffer.offset  = (oskit_u16_t) (&bootplayer);

	pxe_invoke(PXENV_GET_CACHED_INFO, &cacheinfo);
	assert(!cacheinfo.Status);

	/* Stash */
	myip.s_addr       = ((struct in_addr *) &bootplayer.cip)->s_addr;
	serverip.s_addr   = ((struct in_addr *) &bootplayer.sip)->s_addr;
	gatewayip.s_addr  = ((struct in_addr *) &bootplayer.gip)->s_addr;

	/*
	 * Print some useful info.
	 */
	{
		char	cip[16], sip[16], gip[16];

		strcpy(cip, inet_ntoa(myip));
		strcpy(sip, inet_ntoa(serverip));
		strcpy(gip, inet_ntoa(gatewayip));

		printf("DHCP Info: IP: %s, Server: %s, Gateway: %s\n",
		       cip, sip, gip);
	}

	boot_system();

	/*
	 * We don't want to go through the i16 exit since that
	 * waits for a character on the console before reboot!
	 */
	pc_reset();
	return 0;
}

void
boot_system(void)
{
	boot_info_t		boot_info;
	boot_what_t	       *boot_whatp;
	int			success, i, j, err, actual, port;
	struct in_addr		ip;

	err = udp_open();
	if (err)
		goto mbr;

	/*
	 * Try 5 times.
	 */
	success = 0;
	for (i = 0; i < 5; i++) {
		memset(&boot_info, 0, sizeof(boot_info));
		boot_info.opcode = BIOPCODE_BOOTWHAT_REQUEST;

		printf("Sending boot info request %d\n", i);

		err = udp_write(&serverip, BOOTWHAT_SRCPORT, BOOTWHAT_DSTPORT,
				 &boot_info, sizeof(boot_info));
		if (err)
			goto mbr;

		/*
		 * UDP read is nonblocking, so wait for a few seconds.
		 */
		for (j = 0; j < 100; j++) {
			err = udp_read(&myip, BOOTWHAT_SRCPORT,
				       &boot_info, sizeof(boot_info),
				       &actual, &port, &ip);

			if (port != BOOTWHAT_DSTPORT)
				continue;

			if (boot_info.opcode == BIOPCODE_BOOTWHAT_REPLY) {
				success = 1;
				goto gotit;
			}

			delay(100);
		}
	}
 gotit:
	err = udp_close();
	if (err)
		goto mbr;

	if (!success)
		goto mbr;

	/*
	 * Got a seemingly valid boot info packet that says what to do.
	 */
	boot_whatp = (boot_what_t *) &boot_info.data;
	err = 0;
	
	switch (boot_whatp->type) {
	case BIBOOTWHAT_TYPE_MB:
		err = load_multiboot_image(boot_whatp->what.mb.filename);
		break;
		
	case BIBOOTWHAT_TYPE_PART:
		err = boot_mbr_partition(boot_whatp->what.partition);
		break;

	case BIBOOTWHAT_TYPE_SYSID:
		err = boot_mbr_sysid(boot_whatp->what.sysid);

	default:
		PANIC("Invalid type: %d\n", boot_whatp->type);
		break;
	}

	/*
	 * Well, the only way to get here is if the specified boot failed.
	 * Fall through to booting the default, and hope it works.
	 */
 mbr:
#if 1
	printf("Booting the default netboot ...\n");
	load_multiboot_image(NETBOOT);
#else
	printf("Booting the default MBR partition ...\n");
	boot_mbr_default();
#endif
}

/*
 * Assert support that uses way less space
 */
void
local_panic(void)
{
	extern void dump_stack_trace();

	printf("ASSERT FAILED!\n");
	dump_stack_trace();
	boot_mbr_default();
}

/*
 * Get ready to invoke the kernel. Must clean up all the PXE goo.
 */
void
pxe_cleanup(void)
{
	t_PXENV_STOP_UNDI	stop_undi;
	t_PXENV_UNLOAD_STACK	unload_stack;
	t_PXENV_STOP_BASE	stop_base;

	memset(&stop_undi, 0, sizeof(stop_undi));
	memset(&stop_base, 0, sizeof(stop_base));
	memset(&unload_stack, 0, sizeof(unload_stack));

	pxe_invoke(PXENV_STOP_BASE, &stop_base);
	assert(! stop_base.Status);

	pxe_invoke(PXENV_UNLOAD_STACK, &unload_stack);
	assert(! unload_stack.Status);

	pxe_invoke(PXENV_STOP_UNDI, &stop_undi);
	assert(! stop_undi.Status);
}
	
/*
 * This is the first part of a trampoline into the PXE code.
 * Put the processor into 16bit/real mode, and then call the
 * little assembly stub that will invoke the PXE like this:
 *
 *   PXE(our data segment, operation code, argument/return block pointer);
 */
void
pxe_invoke(int routine, void *data)
{
	extern  long pxe_routine;
	extern  long pxe_databucket;
	extern  int  i16_pxe_invoke(void);

	pxe_routine    = routine;
	pxe_databucket = (long) data;
	
	do_16bit(KERNEL_CS, KERNEL_16_CS,
		base_i16_switch_to_real_mode();
		i16_pxe_invoke();
		base_i16_switch_to_pmode();
	);
}

/*
 * We need a trapstate to invoke bios calls.
 */
static struct trap_state ts;

/*
 * Load the Master Boot Record into the magic location (which happens
 * to be _start) using the BIOS readdisk routine.
 */
static void
load_mbr(void)
{
	extern char	_start[];
	int		*foo = (int *) _start;

	DPRINTF("Calling biosread() to overwrite _start ...\n");

	memset(foo, 0, 512);
	biosread(0x80, 0, 0, 1, 1, foo);
}

/*
 * After cleaning up the PXE, switch to 16bit/real mode, and then jump
 * into the MBR.  As long as there is a MBR, and its intact, something
 * good should happen (say, Lilo will appear).
 */
static void
boot_mbr(void)
{
	extern void	i16_boot_mbr(void);

	DPRINTF("Calling pxe_cleanup ...\n");
	pxe_cleanup();

	DPRINTF("Jumping to the MBR ... Bye Bye ...\n");
	do_16bit(KERNEL_CS, KERNEL_16_CS,
		base_i16_switch_to_real_mode();
		i16_boot_mbr();
		base_i16_switch_to_pmode();
	);

	/*
	 * Never returns of course!
	 */
}

/*
 * Simply load and boot the MBR right off the disk.
 */
static int
boot_mbr_default(void)
{
	printf("Loading the MBR and booting the default partition\n");

	load_mbr();
	boot_mbr();
	return 0;
}

/*
 * Boot a particular sysid. Search the table for that partition and
 * make it the active one, then boot. If the type does not exist,
 * fallback to booting the default partition.
 */
int
boot_mbr_sysid(int sysid)
{
	extern char			_start[];
	int				i;
	struct bios_partition_info	*entry;

	printf("Loading the MBR and booting system ID %d\n", sysid);

	load_mbr();

	/*
	 * Look for a non-empty partition that matches the one we want.
	 */
	entry = (struct bios_partition_info *)
		(_start + BIOS_LABEL_BYTE_OFFSET);
		
	for (i = 0; i < 4; i++, entry++) {
		if (!entry->n_sectors)
			continue;
		if (entry->systid == sysid)
			break;
	}

	/*
	 * Either we found it or it does not exist! If we found it,
	 * then make it the bootable partition.
	 */
	if (i < 4) {
		entry = (struct bios_partition_info *)
			(_start + BIOS_LABEL_BYTE_OFFSET);
			
		for (i = 0; i < 4; i++, entry++) {
			if (!entry->n_sectors)
				continue;
			if (entry->systid == sysid)
				entry->bootid = BIOS_BOOTABLE;
			else
				entry->bootid = 0;
		}
	}
	else {
		printf(__FUNCTION__ ": No such partition found: %d\n", sysid);
		return 1;
	}

	boot_mbr();
	return 0;
}

/*
 * Boot a particular partition by making it bootable. Better be non-zero.
 */ 
int
boot_mbr_partition(int partition)
{
	extern char			_start[];
	int				i;
	struct bios_partition_info	*table;

	printf("Loading the MBR and booting system partition %d\n", partition);

	load_mbr();

	table = (struct bios_partition_info *)
		(_start + BIOS_LABEL_BYTE_OFFSET);

	if (! table[partition].n_sectors)
		return 1;

	for (i = 0; i < 4; i++) {
		if (i == partition)
			table[i].bootid = BIOS_BOOTABLE;
		else
			table[i].bootid = 0;
	}

	boot_mbr();
	return 0;
}

/*
 * biosread(dev, cyl, head, sec, nsec, offset)
 *	Read "nsec" sectors from disk to offset "offset" in boot segment
 * BIOS call "INT 0x13 Function 0x2" to read sectors from disk into memory
 *	Call with	%ah = 0x2
 *			%al = number of sectors
 *			%ch = cylinder
 *			%cl = sector
 *			%dh = head
 *			%dl = drive (0x80 for hard disk, 0x0 for floppy disk)
 *			%es:%bx = segment:offset of buffer
 *	Return:		
 *			%al = 0x0 on success; err code on failure
 */
int
biosread(int device, int cyl, int head, int sector, int count, void *addr)
{
	ts.trapno	= 0x13;
	ts.eax		= (0x2 << 8)  | count;
	ts.ecx		= (cyl << 8)  | sector;
	ts.edx		= (head << 8) | device;
	ts.ebx		= (oskit_u32_t) addr;
	ts.v86_es	= 0;

	base_real_int(&ts);
	return ts.eax & 0xFF;
}

/*
 * Kind of a hack.
 */
void
delay(int millis)
{
	
	long long count = millis * 200000;

	while (count)
		count--;
}
